/*
 *      STAGE123...Expected percentage of acceptable lots in single,
 *      double, or triple stage sampling (and the corresponding average
 *      outgoing quality.)
 *
 *               Bob Obenchain, CompuServe User [72007,467]
 *
 *      Parameter settings are either...
 *             specified interactively in response to prompts, or
 *             redirected to come from a batch input file.
 */

#include <stdio.h>
#include <string.h>
#include <process.h>
#include <math.h>

#define abs(a)      (( (a) >= 0 ) ? (a) : (-a))  /* absolute value of a */
#define sgn(a)      (( (a) >= 0 ) ?  1  :  -1 )  /* numerical sign of a */
#define min(a,b)    (( (a) <= (b) ) ?  (a) : (b) )

#define NCONFVAL 99      /* max number of incoming % nonconforming values */

int            stages, ctrlbreak, samsiz1, accnum1, reject1;
int            samsiz2, accnum2, reject2, samsiz3, accnum3;
double         p, q, lp, lq;
double         aoql, plimit, testeff, reweff, ncfm, nclast;
double         accprob[NCONFVAL], nonconfm[NCONFVAL], aoq[NCONFVAL];
FILE           *inpfile, *csvfile;
char           buf[256], inpnam[13], outnam[13];

/* prototypes */

void   main( void );
double logcomb( int n, int k );
double cumprob( int accnum, int samsiz );
double sglprob( int accnum, int samsiz );
double dublsam( int newacc, int newsam, int accept, int reject, int samsiz );
double tripsam( void );

/* Logarithm of the Combinatorial Coefficient */

double logcomb( n, k )
int n, k; {

        int i;
        double  dk, dn, index, logc, log();

        logc = 0;
        if( k >= n || k <= 0 ) return( logc );

        if( n - k < k ) k = n - k;

        dk = (double)k;
        dn = (double)n;

        for( i=0; i<k; i++ ) {
                index = (double)i;
                logc += log( dn - index ) - log( dk - index );
                }

        return( logc );

        }  /* end logcomb */


/* Acceptance Probability Calculation for Single Stage Sampling */

double cumprob( accnum, samsiz )
int accnum, samsiz; {
        int    index;
        double sprob, cprob;
        double exp(), logcomb();

        /* CUMPROB will be set equal to the probability that a BINOMIAL */
        /* random variable with success probability p in each of samsiz */
        /* independent Bernoulli trials will be less than or equal to accnum */
        /* ...where p, q=1-p, lp=log(p), & lq=log(q) are in common storage. */

        cprob = 0;

        if( accnum < 0 || samsiz < 0 || accnum > samsiz ) return( cprob );

        for( index=0; index<=accnum; index++) {

                sprob = exp( logcomb( samsiz, index ) +
                             (double)index * lp +
                             (double)(samsiz-index) * lq );

                if( sprob > 1 )
                        sprob = 1;
                else if( sprob < 0 )
                        sprob = 0;

                cprob += sprob;
                }

        if( cprob > 1 )
               cprob = 1;
        else if( cprob < 0 )
               cprob = 0;

        /* printf( "\nCumprob for %6d, %6d = %6.4lf", accnum, samsiz, cprob); */

        return( cprob );

        }  /* end cumprob */


double sglprob( accnum, samsiz )
int accnum, samsiz; {
        double sprob;
        double exp(), logcomb();

        /* SGLPROB will be set equal to the probability that a BINOMIAL */
        /* random variable with success probability p in each of samsiz */
        /* independent Bernoulli trials will be exactly equal to accnum */
        /* ...where p, q=1-p, lp=log(p), & lq=log(q) are in common storage. */

        sprob = 0;

        if( accnum < 0 || accnum > samsiz ) return( sprob );

        sprob = exp( logcomb( samsiz, accnum ) +
                     (double)accnum * lp +
                     (double)(samsiz-accnum) * lq );

        if( sprob > 1 )
                sprob = 1;
        else if( sprob < 0 )
                sprob = 0;

        /* printf( "\nSglprob for %6d, %6d = %6.4lf", accnum, samsiz, sprob); */

        return( sprob );

        }  /* end sglprob */


/* Acceptance Probability Calculation for Two Stage Sampling */

double dublsam( newacc, newsam, accept, reject, samsiz )
int newacc, newsam, accept, reject, samsiz; {
        int    index;
        double prob1, prob2, probacc, sglprob(), cumprob();

        probacc = cumprob( accept, samsiz );

        for( index=accept+1; index<reject; index++) {
                prob2 = cumprob( newacc-index, newsam );
                prob1 = sglprob( index, samsiz );
                probacc += prob1*prob2;
                }

        if( probacc > 1 )
                 probacc = 1;
        else if( probacc < 0 )
                 probacc = 0;

        /* printf( "\nDublsam for %6d, %6d, %6d, %6d, %6d = %6.4lf",
                newacc, newsam, accept, reject, samsiz, probacc); */

        return( probacc );

        }  /* end dublsam */


/* Acceptance Probability Calculation for Three Stage Sampling */

double tripsam() {
        int index;
        double prob1, prob23, probacc, sglprob(), dublsam();

        probacc = cumprob( accnum1, samsiz1 );

        for( index=accnum1+1; index<reject1; index++) {
                prob23   = dublsam( accnum3-index, samsiz3,
                                    accnum2-index, reject2, samsiz2 );
                prob1    = sglprob( index, samsiz1 );
                probacc += prob1*prob23;
                }

        if( probacc > 1 ) probacc = 1;
        else if( probacc < 0 ) probacc = 0;

        return( probacc );

        }  /* end tripsam */

void main() {

        char   kb;
        int    index;
        double cumprob(), dublsam(), tripsam();
        extern double log();
        extern FILE *fopen();                   /* Stream I/O */
        extern int fclose();                    /* Stream Output */

        printf("\n\n*** One, Two, or Three Stage Acceptance Sampling ***");
        printf("\n        Operating Characteristic Calculations");
        printf("\n            and Average Outgoing Quality.\n");
        printf("\n\n             Stage123.EXE...Version 9310");
        printf("\n\n          A Quality Assurance Training Tool:");
        printf("\n   Statistics Committee of the QA Section of the PMA.");
        printf("\n\n      Bob Obenchain, CompuServe User [72007,467]\n\n" );

        printf(      "\tWill Parameter Settings be Input via K = Keyboard ? ...or" );
        printf(    "\n\t                                     B = Batch File ?" );
        printf(  "\n\n\tENTER a   K   or   B   keypress now --> " );
        while( 1 ) {
              gets( buf );
              sscanf( buf, "%c", &kb );
              if( kb == 'k' )
                  kb  = 'K';
              if( kb == 'b' )
                  kb  = 'B';
              if( kb == 'K' || kb == 'B' )
                  break;
              printf( "\r\tENTER either K  or  B  ...Try Again --> " );
              }

        strcpy( outnam, "stage123" );
        strcpy( inpnam, "stage123" );

        if( kb == 'B' )
            printf( "\n\n\tBatch Input of Parameter Settings Selected..." );
        else
            printf( "\n\n\tKeyboard Input of Parameter Settings Selected..." );

        printf( "\n\n\tAt colon Prompts : ...simply press ENTER to get the [default]." );

        if( kb == 'B' ) {
            printf(  "\n\n\tSpecify filename of Batch Input File [%s] : ", inpnam );
            gets( buf );
            if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                sscanf( buf, "%s", inpnam );
            inpnam[ min( 8, (int)strcspn( inpnam, ". " ) ) ] = '\0';
            if( inpnam[0] == '\0' )
                strcpy( inpnam, "stage123" );
            strcpy( outnam, inpnam );
            strcat( inpnam, ".in" );
            printf( "\n\tThe Batch Input file is to be: %s\n", inpnam );
            if( ( inpfile = fopen( inpnam, "r" ) ) == NULL ) {
                    printf(  "\tCannot read Batch Input file: %s\n", inpnam );
                    printf(  "\t...using Keyboard Input from standard Infile, stdin.\n" );
                    kb = 'K';
                    }
            }

        if( kb == 'K' )
            inpfile = stdin;

        printf(  "\n\n\tSpecify filename to save primary Stage123 Output [%s] : ",
            outnam );
        fgets( buf, 80, inpfile );
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf( buf, "%s", outnam );
        outnam[ min( 8, (int)strcspn( outnam, ". " ) ) ] = '\0';
        if( outnam[0] == '\0' )
            strcpy( outnam, "stage123" );
        strcat( outnam, ".csv" );
        printf(  "\n\tThe Stage123 primary Output file is to be: %s\n",
                outnam );
        if( ( csvfile = fopen( outnam, "w" ) ) == NULL ) {
            printf(  "\tCannot write to Output filename : %s\n", outnam );
            if( ( csvfile = fopen( "stage123.csv", "w" ) ) == NULL )
                    printf(  "\tCannot write to Output filename : stage123.csv\n" );
            else {
                    printf(  "\t...using default Outfile name : stage123.csv\n" );
                    strcpy( outnam, "stage123.csv" );
                    }
            }

        stages = 1;

        samsiz1 = 50;
        printf("\n\nEnter First Stage Sample Size [%d] : ", samsiz1 );
        fgets(buf, 100, inpfile);
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf(buf, "%d", &samsiz1);
        if( samsiz1 <= 0 ) {
                printf("\nFirst Stage Sample Size will be 50, the default.\n");
                samsiz1 = 50;
                }
        else if( samsiz1 == 1 ) {
                printf("\nFirst Stage Sample Size will be 2, the minimum allowed.\n");
                samsiz1 = 2;
                }
        else if( samsiz1 > 5000 ) {
                printf("\nFirst Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                samsiz1 = 5000;
                }

        accnum1 = 0;
        printf("\nEnter First Stage Accept Number [%d] : ", accnum1 );
        fgets(buf, 100, inpfile);
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf(buf, "%d", &accnum1);
        if( accnum1 < -1 ) {
                printf("\nStage 1 Accept Number will be 0, the default.\n");
                accnum1 = 0;
                }
        else if( accnum1 == -1 ) {
                printf("\nA Stage 1 Accept Number of -1 indicates that the lot cannot be accepted\n");
                printf("at this first stage even if zero nonconformances are observed.\n");
                }
        else if( accnum1 >= samsiz1 ) {
                accnum1 = samsiz1/4;
                printf("\nStage 1 Accept Number will be %d, which is about one-fourth the Sample Size of %d.\n", accnum1, samsiz1);
                }
        if( accnum1 > 999 ) {
                printf("\nAccept Numbers cannot exceed 999.\n");
                accnum1 = 999;
                }

        reject1 = accnum1 + 1;
        printf("\nEnter First Stage Reject Number [%d] : ", reject1 );
        fgets(buf, 100, inpfile);
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf(buf, "%d", &reject1);
        if( reject1 < 1 )
            reject1 = 1;
        if( reject1 <= accnum1 + 1 ) {
                if( accnum1 < 0 ) {
                        printf("\nAccept Number reset to Zero for Single Stage Sampling.\n");
                        accnum1 = 0;
                        }
                reject1 = accnum1 + 1;
                if( reject1 > samsiz1 ) {
                        reject1 = samsiz1;
                        printf("\nStage 1 Reject Number will be %d, ...all sampled units nonconform.\n", reject1);
                        }
                else {
                        printf("\nA Reject Number of %d is the minimum allowed for\n", reject1);
                        printf("Single Stage Sampling with Accept Number of %d.\n", accnum1);
                        }
                }

        if( reject1 > accnum1 + 1 ) {   /* Second Stage Required */

                stages = 2;

                samsiz2 = samsiz1;
                printf("\nEnter Second Stage Sample Size [%d] : ", samsiz2 );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%d", &samsiz2);
                if( samsiz2 <= 0 ) {
                        printf("\nSecond Stage Sample Size will be 50, the default.\n");
                        samsiz2 = 50;
                        }
                else if( samsiz2 > 5000 ) {
                        printf("\nSecond Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz2 = 5000;
                        }

                accnum2 = reject1 - 1;
                printf("\nEnter Second Stage Accept Number [%d] : ", accnum2 );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%d", &accnum2);
                if( accnum2 < accnum1 ) {
                        accnum2 = accnum1;
                        printf("\nSecond Stage Accept Number will be %d, which is the minimum allowed when\n", accnum2);
                        printf("the First Stage Accept Number is %d.\n", accnum1);
                        }
                if( accnum2 < reject1 - 1 ) {
                        printf("\nA Third Stage is required when the Second Stage Accept Number is");
                        printf("\ntwo or more units smaller than the First Stage Reject Number.");
                        }
                if( accnum2 >= samsiz1 + samsiz2 ) {
                        accnum2 = (samsiz1 + samsiz2)/4;
                        printf("\nSecond Stage Accept Number will be %d, which is about one-fourth of the", accnum2);
                        printf("\n1st and 2nd Stage Sample Size of %d.\n", samsiz1+samsiz2);
                        }
                if( accnum2 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum2 = 999;
                        }

                reject2 = accnum2 + 1;
                printf("\nEnter Second Stage Reject Number [%d] : ", reject2 );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%d", &reject2);
                if( reject2 <= accnum2 + 1 ) {
                        reject2 = accnum2 + 1;
                        if( reject2 > samsiz1 + samsiz2 ) {
                                reject1 = samsiz1 + samsiz2;
                                printf("\nStage 2 Reject Number will be %d, ...all sampled units nonconform.\n", reject2);
                                }
                        else {
                                printf("\nA Reject Number of %d is the minimum allowed for\n", reject2);
                                printf("Two Stage Sampling with Accept Number of %d.\n", accnum2);
                                }
                        }
                }

        if( reject2 > accnum2 + 1 ) {   /* Third Stage Required */

                stages = 3;

                samsiz3 = samsiz2;
                printf("\nEnter Third Stage Sample Size [%d] : ", samsiz3 );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%d", &samsiz3);
                if( samsiz3 <= 0 ) {
                        printf("\nThird Stage Sample Size will be 50, the default.\n");
                        samsiz3 = 50;
                        }
                else if( samsiz3 > 5000 ) {
                        printf("\nThird Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz3 = 5000;
                        }

                accnum3 = reject2 - 1;
                printf("\nEnter Third Stage Accept Number [%d] : ", accnum3 );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%d", &accnum3);
                if( accnum3 < reject2 - 1 ) {
                        accnum3 = reject2 - 1;
                        printf("\nThe Third Stage Accept Number will be %d, which is the minimum allowed when\n", accnum3);
                        printf("the Second Stage Reject Number is %d.\n", reject2);
                        }
                if( accnum3 >= samsiz1 + samsiz2 + samsiz3 ) {
                        accnum2 = (samsiz1 + samsiz2 + samsiz3)/4;
                        printf("\nThird Stage Accept Number will be %d, which is about one-fourth of the", accnum3);
                        printf("\nThree Stage Sample Total Size of %d.\n", samsiz1+samsiz2+samsiz3);
                        }
                if( accnum3 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum3 = 999;
                        }
                }

        testeff = 100.0;
        printf("\nSpecify Testing Efficiency Percentage [%5.1lf] : ", testeff );
        fgets(buf, 100, inpfile);
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf(buf, "%lf", &testeff);
        testeff /= 100;
        if( testeff > 1 || testeff < 0.01 ) {
                printf("\nTesting Efficiency will be 100%%, the default.\n");
                testeff = 1;
                }
        else if( testeff < 0.1 ) {
                printf("\nTesting Efficiency will be 10%%, the minimum allowed here.\n");
                testeff = 0.1;
                }

        reweff = 100.0;
        printf("\n\nSpecify Re-Work Efficiency Percentage [%5.1lf] : ", reweff );
        fgets(buf, 100, inpfile);
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf(buf, "%lf", &reweff);
        reweff /= 100;
        if( reweff > 1 || reweff < 0.01 ) {
                printf("\nRe-Work Efficiency will be 100%%, the default.\n");
                reweff = 1;
                }
        else if( reweff < 0.1 ) {
                printf("\nRe-Work Efficiency will be 10%%, the minimum allowed here.\n");
                reweff = 0.1;
                }

        if( stages == 1 )
                printf("\n\nSUMMARY: One Stage Sampling...\n");
        else if( stages == 2 )
                printf("\n\nSUMMARY: Two Stage Sampling...\n");
        else
                printf("\n\nSUMMARY: Three Stage Sampling...\n");

        printf("\nFirst Stage Sample Size = %d", samsiz1 );
        fprintf( csvfile, "%4d, = First Stage Sample Size,", samsiz1 );
        printf("\nFirst Stage Accept Number = %d", accnum1 );
        fprintf( csvfile, "\n%4d, = First Stage Accept Number,", accnum1 );
        printf("\nFirst Stage Reject Number = %d", reject1 );
        fprintf( csvfile, "\n%4d, = First Stage Reject Number,", reject1 );

        if( stages >= 2 ) {
                printf("\nSecond Stage Sample Size = %d", samsiz2 );
                fprintf( csvfile,"\n%4d, = Second Stage Sample Size,", samsiz2 );
                printf("\nSecond Stage Accept Number = %d", accnum2 );
                fprintf( csvfile,"\n%4d, = Second Stage Accept Number,", accnum2 );
                printf("\nSecond Stage Reject Number = %d", reject2 );
                fprintf( csvfile,"\n%4d, = Second Stage Reject Number,", reject2 );
                }

        if( stages == 3 ) {
                printf("\nThird Stage Sample Size = %d", samsiz3 );
                fprintf( csvfile,"\n%4d, = Third Stage Sample Size,", samsiz3 );
                printf("\nFinal Accept Number = %d", accnum3 );
                fprintf( csvfile,"\n%4d, = Final Accept Number,", accnum3 );
                }

        printf("\nTesting Efficiency = %6.1lf%%", 100*testeff );
        fprintf( csvfile, "\n%6.1lf, = Testing Efficiency %%,", 100*testeff );

        printf("\nRe-Work Efficiency = %6.1lf%%", 100*reweff );
        fprintf( csvfile, "\n%6.1lf, = Re-Work Efficiency %%,", 100*reweff );

        printf("\n\nPress ENTER to start calculations...");
        fprintf( csvfile, "\nIn%%NConf,%%Yield,Out%%NConf");
        if( kb == 'K' )
            getchar();

        aoql = plimit = 0;
        nclast = -1;
        ncfm = 1.0;

        for(index=0; index<NCONFVAL; index++) {

                printf("\n\nStep Number = %d", index);

                printf("\n\nEnter a Nonconformance Percentage [%6.2lf] : ", ncfm );
                fgets(buf, 100, inpfile);
                if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                    sscanf(buf, "%lf", &ncfm);

                if( ncfm < 0 ) break;

                if( ncfm > 100 ) {
                        printf("\nNonconformance Percentage will be 1%%, the default.\n");
                        ncfm = 1.0;
                        }

                nonconfm[index] = ncfm;

                p = testeff*ncfm/100;  /* convert % to fraction */

                if( p == 0 )
                        accprob[index] = 1;

                else if( p == 1 )
                        accprob[index] = 0;

                else if( ncfm != nclast ) {

                        /* perform calculations only when the incoming
                           nonconformance percentage is > 0, < 100, and
                           has changed from the last step, nclast. */

                        lp = log( p );
                        q = 1 - p;
                        lq = log( q );

                        if( stages == 1 )
                                accprob[index] = cumprob( accnum1, samsiz1 );

                        else if( stages == 2 )
                                accprob[index] = dublsam( accnum2, samsiz2,
                                                 accnum1, reject1, samsiz1 );
                        else if( stages == 3 )
                                accprob[index] = tripsam();
                        }

                else accprob[index] = accprob[index-1];

                aoq[index] = ( accprob[index] + (1-reweff)*(1-accprob[index]) ) * ncfm;

                if( aoq[index] > aoql ) {

                        aoql = aoq[index];

                        plimit = ncfm;

                        }

                printf("\n\n For Incoming Nonconformance Percentage = %6.2lf,",
                        ncfm);
                printf("\n Expected First-Pass Yield Percentage   = %6.2lf,",
                        100*accprob[index]);
                printf("\n and Average Outgoing %% Nonconforming   = %6.2lf.\n\n",
                        aoq[index]);

                if( ncfm != nclast ) {

                        /* Successive duplicates not writtem to output file. */

                        fprintf( csvfile, "\n %6.2lf,%6.2lf,%6.2lf",
                                 ncfm, 100*accprob[index], aoq[index]);
                        }

                nclast = ncfm;

                printf("\n\nTo QUIT, enter a NEGATIVE Nonconformance Percentage...");
                }

        printf( "\n At the worst calculated incoming %%NConf of %6.2lf,", plimit);
        printf( "\n the Lower Bound for AOQL %%NConf is %6.2lf.\n\n", aoql);

        fprintf( csvfile, "\n-1,...End of Input Signal...,");
        fprintf( csvfile, "\n At the worst calculated incoming %%NConf of %6.2lf,,", plimit);
        fprintf( csvfile, "\n the Lower Bound for AOQL, %%NConf is, %6.2lf.\n\n", aoql);
        fclose(csvfile);

        exit(0);
        }

